"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const DiscordClient_1 = __importDefault(require("./DiscordClient"));
const CommandHandler_1 = __importDefault(require("./CommandHandler"));
const Enums_1 = require("../consts/Enums");
const boltzrpc_pb_1 = require("../proto/boltzrpc_pb");
const BoltzClient_1 = require("../boltz/BoltzClient");
const Utils_1 = require("../Utils");
class NotificationProvider {
    constructor(logger, service, boltz, config, currencies) {
        this.logger = logger;
        this.service = service;
        this.boltz = boltz;
        this.config = config;
        this.currencies = currencies;
        this.backendName = 'backend';
        // These Sets contain the symbols for which an alert notification was sent
        this.walletAlerts = new Set();
        this.localBalanceAlerts = new Set();
        this.remoteBalanceAlerts = new Set();
        this.disconnected = new Set();
        this.init = () => __awaiter(this, void 0, void 0, function* () {
            try {
                yield this.discord.init();
                yield this.discord.sendMessage('Started Boltz instance');
                this.logger.verbose('Connected to Discord');
                const check = () => __awaiter(this, void 0, void 0, function* () {
                    yield this.checkConnections();
                    yield this.checkBalances();
                });
                yield check();
                this.logger.debug(`Checking balances and connection status every ${this.config.interval} minutes`);
                this.timer = setInterval(() => __awaiter(this, void 0, void 0, function* () {
                    yield check();
                }), Utils_1.minutesToMilliseconds(this.config.interval));
            }
            catch (error) {
                this.logger.warn(`Could not connect to Discord: ${error}`);
            }
        });
        this.disconnect = () => {
            clearInterval(this.timer);
        };
        this.checkConnections = () => __awaiter(this, void 0, void 0, function* () {
            const info = yield this.boltz.getInfo();
            info.chainsMap.forEach(([symbol, chain]) => __awaiter(this, void 0, void 0, function* () {
                yield this.checkConnection(`${symbol} node`, chain.chain);
                yield this.checkConnection(`${symbol} LND`, chain.lnd);
            }));
        });
        this.checkConnection = (service, object) => __awaiter(this, void 0, void 0, function* () {
            if (object) {
                if (object.error === '') {
                    if (this.disconnected.has(service)) {
                        this.disconnected.delete(service);
                        yield this.sendReconnected(service);
                    }
                    return;
                }
            }
            if (!this.disconnected.has(service)) {
                this.disconnected.add(service);
                yield this.sendLostConnection(service);
            }
        });
        this.checkBalances = () => __awaiter(this, void 0, void 0, function* () {
            const balances = yield Utils_1.parseBalances(yield this.boltz.getBalance());
            for (const currency of this.currencies) {
                const balance = balances.get(currency.symbol);
                if (balance) {
                    const { symbol, minWalletBalance, minLocalBalance, minRemoteBalance } = currency;
                    yield this.checkBalance(symbol, this.walletAlerts, balance.walletBalance.totalBalance, minWalletBalance, true);
                    if (balance.lightningBalance) {
                        const { localBalance, remoteBalance } = balance.lightningBalance;
                        yield this.checkBalance(symbol, this.localBalanceAlerts, localBalance, minLocalBalance, false, true);
                        yield this.checkBalance(symbol, this.remoteBalanceAlerts, remoteBalance, minRemoteBalance, false, false);
                    }
                }
            }
        });
        this.checkBalance = (currency, set, balance, threshold, isWallet, isLocal) => __awaiter(this, void 0, void 0, function* () {
            const sentAlert = set.has(currency);
            if (sentAlert) {
                if (balance > threshold) {
                    set.delete(currency);
                    yield this.sendRelief(currency, balance, threshold, isWallet, isLocal);
                }
            }
            else {
                if (balance <= threshold) {
                    set.add(currency);
                    yield this.sendAlert(currency, balance, threshold, isWallet, isLocal);
                }
            }
        });
        this.listenToDiscord = () => {
            this.discord.on('error', (error) => {
                this.logger.warn(`Discord client threw: ${error.message}`);
            });
        };
        this.listenToBoltz = () => {
            this.boltz.on('status.updated', (status) => __awaiter(this, void 0, void 0, function* () {
                switch (status) {
                    case BoltzClient_1.ConnectionStatus.Connected:
                        if (this.disconnected.has(this.backendName)) {
                            this.disconnected.delete(this.backendName);
                            yield this.sendReconnected(this.backendName);
                        }
                        break;
                    case BoltzClient_1.ConnectionStatus.Disconnected:
                        if (!this.disconnected.has(this.backendName)) {
                            this.disconnected.add(this.backendName);
                            yield this.sendLostConnection(this.backendName);
                        }
                        break;
                }
            }));
        };
        this.listenToService = () => {
            this.service.on('swap.successful', (swap) => __awaiter(this, void 0, void 0, function* () {
                yield this.sendSwapSuccessful(swap);
            }));
        };
        this.sendAlert = (currency, balance, threshold, isWallet, isLocal) => __awaiter(this, void 0, void 0, function* () {
            const { actual, expected } = this.formatBalances(balance, threshold);
            const missing = Utils_1.satoshisToCoins(threshold - balance);
            const balanceName = this.getBalanceName(isWallet, isLocal);
            this.logger.warn(`${currency} ${balanceName} balance is less than ${threshold}: ${balance}`);
            // tslint:disable-next-line:prefer-template
            let message = ':rotating_light: **Alert** :rotating_light:\n\n' +
                `The ${currency} ${balanceName} balance of ${actual} ${currency} is less than expected ${expected} ${currency}\n\n` +
                `Funds missing: **${missing} ${currency}**`;
            if (isWallet) {
                const { address } = yield this.boltz.newAddress(currency, boltzrpc_pb_1.OutputType.COMPATIBILITY);
                message += `\nDeposit address: **${address}**`;
            }
            yield this.discord.sendMessage(message);
        });
        this.sendRelief = (currency, balance, threshold, isWallet, isLocal) => __awaiter(this, void 0, void 0, function* () {
            const { actual, expected } = this.formatBalances(balance, threshold);
            const balanceName = this.getBalanceName(isWallet, isLocal);
            this.logger.info(`${currency} ${balanceName} balance is more than expected ${threshold} again: ${balance}`);
            yield this.discord.sendMessage(`The ${currency} ${balanceName} balance of ${actual} ${currency} is more than expected ${expected} ${currency} again`);
        });
        this.sendSwapSuccessful = (swap) => __awaiter(this, void 0, void 0, function* () {
            const isReverse = swap.status === Enums_1.SwapUpdateEvent.InvoiceSettled;
            const feeSymbol = Utils_1.getFeeSymbol(swap.pair, swap.orderSide, isReverse);
            const getSwapDirection = () => {
                let { base, quote } = Utils_1.splitPairId(swap.pair);
                // Switch the symbols if the swap was a sell order
                if (swap.orderSide === boltzrpc_pb_1.OrderSide.SELL) {
                    [base, quote] = [quote, base];
                }
                let direction;
                if (isReverse) {
                    direction = `Lightning ${quote} to onchain ${base}`;
                }
                else {
                    direction = `onchain ${quote} to Lightning ${base}`;
                }
                return direction;
            };
            const message = `Swapped ${getSwapDirection()} and earned ${Utils_1.satoshisToCoins(swap.fee)} ${feeSymbol} in fees`;
            yield this.discord.sendMessage(message);
        });
        this.sendLostConnection = (service) => __awaiter(this, void 0, void 0, function* () {
            yield this.discord.sendMessage(`**Lost connection to ${service}**`);
        });
        this.sendReconnected = (service) => __awaiter(this, void 0, void 0, function* () {
            yield this.discord.sendMessage(`Reconnected to ${service}`);
        });
        this.formatBalances = (balance, threshold) => {
            return {
                actual: Utils_1.satoshisToCoins(balance),
                expected: Utils_1.satoshisToCoins(threshold),
            };
        };
        this.getBalanceName = (isWallet, isLocal) => {
            if (isWallet) {
                return 'wallet';
            }
            else {
                return isLocal ? 'local' : 'remote';
            }
        };
        this.discord = new DiscordClient_1.default(config.token, config.channel, config.prefix);
        this.listenToBoltz();
        this.listenToDiscord();
        this.listenToService();
        new CommandHandler_1.default(this.logger, this.service, this.boltz, this.discord);
    }
}
exports.default = NotificationProvider;
//# sourceMappingURL=NotificationProvider.js.map